// -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; indent-tabs-mode: nil; -*-
//////////////////////////////////////////////////////////////////////
// File: timedgate2.cpp
// This file contains sample implementation for the test generation:
// Timed train gate controller implementation under test.
// Authors:
//   Marius Mikucionis marius@cs.aau.dk
//////////////////////////////////////////////////////////////////////

#include "timedgate2.h"
#include "tron/timeutil.h"

#include "assert.h"
#include <iostream>
#include <stdio.h>
#include <sys/errno.h>
#include <string.h>
#include <stdlib.h>

using std::cout; // for stateDump
using std::cerr;
using std::endl;
using std::queue;


int instances=0;
extern "C" {
    TestAdapter* adapter_new(Reporter* rep, int argc, const char* args[]){
        if (instances>0) {
            cerr << "Only one instance at a time is supported. "
                "Destroy the old adapter and retry" << endl;
            return NULL;
        } else instances++;
        int m = 0;
        if (argc>1) m = atoi(args[1]);
        if (0<=m && m<=6) return new GateAdapter(rep, 4, m);
        else {
            cerr << "Mutant "<< m << " ("<<args[1]<<") not supported" << endl;
            return NULL;
        }
    };
    void adapter_delete(TestAdapter* ptr){ delete ptr; instances--; };
    void adapter_start(TestAdapter* a){ ((GateAdapter*)a)->start_imp(); };
    void adapter_perform(TestAdapter* a,
                         int32_t chan, uint16_t n, const int32_t data[]) {
        ((GateAdapter*)a)->perform_imp(chan, n, data);
    }
}


const uint16_t inCount = 2, outCount = 2;
const char* const inStr[inCount]={"appr", "leave"};
const char* const outStr[outCount]={"go", "stop"};
int32_t ins[inCount], outs[outCount], release=-1;
uint16_t appr=0, leave=1, go=0, stop=1;
int32_t timeUnit=10000;

GateAdapter::GateAdapter(Reporter* rep, int railways, int mutant):
    TestAdapter(rep), alive(false), abort(false),
    gatecnt(railways)
{
    start = adapter_start;
    perform = adapter_perform;
    assert(gatecnt>0);
    if (0 != tron_mutex_init(&input_m, NULL/*use defaults*/)) {
        cerr << "GA: error while init mutex" << endl;
        exit(1);
    }
    tron_cond_init(&input_c, NULL);
    tron_cond_init(&ack_c, NULL);

    try { // initialize testing interface
        int r = 0;
        for (int i = 0; i < inCount; i++) {
            r = rep->getInputEncoding(rep, inStr[i]);
            if (r<0) throw r;
            ins[i] = r;
        }
        for (int i = 0; i < outCount; i++) {
            r = rep->getOutputEncoding(rep, outStr[i]);
            if (r<0) throw r;
            outs[i] = r;
        }
        r = rep->addVarToInput(rep, ins[appr], "envTrain"); if (r<0) throw r;
        r = rep->addVarToInput(rep, ins[leave], "envTrain"); if (r<0) throw r;
        r = rep->addVarToOutput(rep, outs[go], "envTrain"); if (r<0) throw r;
        r = rep->addVarToOutput(rep, outs[stop], "envTrain"); if (r<0) throw r;
        r = rep->setTimeUnit(rep, timeUnit); if (r<0) throw r;
        r = rep->setTimeout(rep, 100000); if (r<0) throw r;
    } catch (int error) {
        cerr << "GA: "<< rep->getErrorMessage(rep, error) << endl;
        exit(EXIT_FAILURE);
    }
}

GateAdapter::~GateAdapter()
{
    {        
        TronLock lock(input_m);
        abort = true; 
        tron_cond_broadcast(&input_c);
    }
    joinThread(NULL); // important to wait for thread exit

    tron_cond_destroy(&input_c);
    tron_cond_destroy(&ack_c);
    tron_mutex_destroy(&input_m);
}

inline void GateAdapter::start_imp()
{
    {
        TronLock lock(input_m);
        startonce();
        while (!alive) tron_cond_wait(&ack_c, &input_m);
    }
}

inline void GateAdapter::perform_imp(int32_t chan, uint16_t n,
                                 const int32_t data[])
{
    {
        TronLock lock(input_m);
        inpChan.push(chan);
        if (n>0) inpTrain.push(data[0]);
        else inpTrain.push(-1);
        tron_cond_broadcast(&input_c);
    }
}

void* GateAdapter::threadExecute()
{
    int32_t chan, train, id;
    bool opened = true;
    std::queue<int32_t> trains; // queue of trains
    { // initialize and notify that we are up:
        TronLock lock(input_m);
        alive = true; tron_cond_broadcast(&ack_c);
    }        
    while (true) {
        { // wait for inputs:
            TronLock lock(input_m);
            while (inpChan.empty() && !abort)
                tron_cond_wait(&input_c, &input_m);
            if (abort) break;
            chan = inpChan.front(); inpChan.pop();
            train = inpTrain.front(); inpTrain.pop();
        }
        // react to inputs:
        if (chan == ins[appr]) {
            trains.push(train);
            if (opened) {
                opened = false;
            } else {
                reportStop(train);
            }
        } else if (chan == ins[leave]) {
            id = trains.front(); trains.pop();
            if (id != train) {
                cerr << "GA: different trains: front=" << id << " leave=" 
                     << train << endl;
            }
            assert(id == train);
            if (trains.empty()) {
                opened = true;
            } else {
                train = trains.front();
                reportGo(train);
            }
        } else {
            cerr << "GA: got bad input " << chan << endl;
        }
    }
    if (!trains.empty()) {
        cout << "GA: trains in queue: ";
        do {
            cout << trains.front() << " ";
            trains.pop();
        } while (!trains.empty());
        cout << endl;
    }
}

inline void GateAdapter::reportGo(const int32_t &train)
{
    rep->report_now(rep, outs[go], 1, &train);
}

inline void GateAdapter::reportStop(const int32_t &train)
{
    rep->report_now(rep, outs[stop], 1, &train);
}
